//------------------------------------------------------------------------------
// saner.cpp - модуль лексического анализатора
//------------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "scanner_data.h"
#include "scanner_local.h"

//------------------------------------------------------------------------------

void er(int i); // Импорт интерфейса функции вывода ошибок

//------------------------------------------------------------------------------
// Функции транслитераторра, используемые для определения класса лексем
//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу букв
bool inline isLetter(int ch) {
  if((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу двоичных цифр
bool inline isBin(int ch) {
  if((ch == '0' || ch == '1'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу восьмиричных цифр
bool inline isOctal(int ch) {
  if((ch >= '0' && ch <= '7'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу десятичных цифр
bool inline isDigit(int ch) {
  if((ch >= '0' && ch <= '9'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу шестнадцатиричных цифр
bool inline isHex(int ch) {
  if((ch >= '0' && ch <= '9') ||
     (ch >= 'A' && ch <= 'F') || 
     (ch >= 'a' && ch <= 'f'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу пробелов
bool inline isSkip(int ch) {
  if(ch == ' ' || ch == '\t' || ch == '\n' || ch == '\f')
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу игнорируемых
bool inline isIgnore(int ch) {
  if(ch >0 && ch < ' ' && ch != '\t' && ch != '\n' && ch != '\f')
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Читает следующий символ из входного потока
static void nxch(void)
{
    if((ch = getc(infil)) == '\n') {++line; column = 0;}
    else ++column;
}

//------------------------------------------------------------------------------
// Максимальное значение двоичного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uBinMax = 0x3FFFFFFF; // 30 двоичных разрядов
//------------------------------------------------------------------------------

// Преобразование строки двоичных цифр в целое число
unsigned int binToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    unsigned int digit = static_cast<unsigned int>(*str);
    if(uBinMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    num <<= 1;
    num += digit - static_cast<unsigned int>('0');
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------
// Максимальное значение восьмеричного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uOctMax = 0x0FFFFFFF; // 28 двоичных разрядов
//------------------------------------------------------------------------------

// Преобразование строки восьмеричных цифр в целое число
unsigned int octToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    unsigned int digit = static_cast<unsigned int>(*str);
    if(uOctMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    num = (num << 3) + (digit - static_cast<unsigned int>('0'));
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------
// Максимальное значение двоичного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uDecMax = (0x7FFFFFFF / 10); // деление на 10 максимального целого
//------------------------------------------------------------------------------

// Преобразование строки десятичных цифр в целое число
unsigned int decToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    unsigned int digit = static_cast<unsigned int>(*str);
    if(uDecMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    num = (num * 10) + (digit - static_cast<unsigned int>('0'));
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------
// Максимальное значение восьмеричного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uHexMax = 0x07FFFFFF; // 28 двоичных разрядов
//------------------------------------------------------------------------------

// Преобразование строки шестнадцатиричных цифр в целое число
unsigned int hexToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    if(uDecMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    unsigned int digit = static_cast<unsigned int>(*str);
    num = num << 4;
    if(*str >= '0' && *str <= '9')
    {
      num += (digit - static_cast<unsigned int>('0'));
    }
    else if(*str >= 'A' && *str <= 'F')
    {
      num += (digit - static_cast<unsigned int>('A')) + 10;
    }
    else //(*str >= 'a' && *str <= 'f')
    {
      num += (digit - static_cast<unsigned int>('a')) + 10;
    }
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------

// Определяет принадлежность к ключевому слову
static void FindKw(void)
{
  for(int i = 0; i < rw_size; ++i)
    if(!strcmp(rwtab[i].wrd, lv)) {
      lc = rwtab[i].lc;
      return;
    }
  lc = lexId;
}

//------------------------------------------------------------------------------

// Проверка на идентификатор или ключевое слово
static void IsContinueIdOrKw(void)
{
    while(isLetter(ch) || isDigit(ch) || ch=='_') {
        lv[++i_lv] = ch;
        nxch();
    }
    lv[++i_lv] = '\0';
    FindKw(); 
}

//------------------------------------------------------------------------------

// Знак деления или комментарий
static void IsSlashComment(void)
{
    if(ch != '*') {lc = lexSlash; return;}
    nxch();
loop: 
    if(ch == EOF) { 
      lc = lexError; lv[++i_lv]='\0'; er(2); return; 
    } 
    while(ch != '*') {
        //lv[++i_lv] = ch;
        if(ch == EOF) {lc = lexError; lv[++i_lv]='\0'; er(2); return;} 
        nxch();
    }
    nxch(); if(ch != '/') 
    {
      //lv[++i_lv] = ch; 
      goto loop;
    }
    lv[++i_lv] = '\0';
    lc = lexComment;
    nxch();
}

//------------------------------------------------------------------------------

// Проверка порядка числа
static void IsExp(void) {
  lv[++i_lv]=ch; nxch();
  //_0:
    if(ch == '+' || ch == '-') {lv[++i_lv]=ch; nxch(); goto _1;}
    if(isDigit(ch)) {lv[++i_lv]=ch; nxch(); goto _2;}
    lc=lexError; lv[++i_lv]='\0'; return;
  _1:
    if(isDigit(ch)) {lv[++i_lv]=ch; nxch(); goto _2;}
    lc=lexError; lv[++i_lv]='\0'; er(3); return;
  _2:
    if(isDigit(ch)) {lv[++i_lv]=ch; nxch(); goto _2;}
    if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return;}
    lc=lexFloat; lv[++i_lv]='\0'; 
    dnum = atof(lv);
    return;
}

//------------------------------------------------------------------------------

// Число с плавающей точкой
static void IsFltNumContinue(void) {
  lv[++i_lv]=ch; nxch();
  _0:
    if(isDigit(ch)) {lv[++i_lv]=ch; nxch(); goto _0;}
    if(ch == 'E' || ch == 'e') {
      IsExp(); 
      return;
    }
    if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return;}
    lc=lexFloat; lv[++i_lv]='\0'; 
    dnum = atof(lv);
    return; 
}

//------------------------------------------------------------------------------

// Целое или действительное
static void IsNumber(void) {
  do {
    lv[++i_lv] = ch;
    nxch();
  } while (isDigit(ch));
  if(ch == '.') {IsFltNumContinue(); return;}
  if(ch=='e'||ch=='E') {IsExp();return;}
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return;}
  lc = lexInt; lv[++i_lv] = '\0'; 
  unum = decToInt(lv);
}

//------------------------------------------------------------------------------

// Целое с префиксом
static void IsPrefNumber(void) {
    //_0:
        if(isDigit(ch)) {
          lv[++i_lv] = ch; nxch(); goto _1;
        }
        lc = lexError; lv[++i_lv] = '\0'; er(3); return;
    _1:
        if(isDigit(ch)) {lv[++i_lv] = ch; nxch(); goto _1;}
        if(ch=='}') {
          lv[++i_lv] = '\0'; nxch();
          if(!strcmp(lv, "2")) {lv[i_lv]='#'; goto _2;}
          if(!strcmp(lv, "8")) {lv[i_lv]='#'; goto _3;}
          if(!strcmp(lv, "10")) {lv[i_lv]='#'; goto _4;}
          if(!strcmp(lv, "16")) {lv[i_lv]='#'; goto _5;}
          lc = lexError; lv[++i_lv] = '\0'; er(4); return;
        }
        lc = lexError; lv[++i_lv] = '\0'; er(3); return;
    _2:
        if(isBin(ch)) {lv[++i_lv]=ch; nxch(); goto _2;}
        if(isLetter(ch) || isDigit(ch)) 
        {
          lc = lexError; lv[++i_lv]='\0'; er(3); return;
        }
        lc = lexInt; lv[++i_lv] ='\0'; 
        unum = binToInt(lv + 2);
        return;
    _3:
        if(isOctal(ch)) {lv[++i_lv]=ch; nxch(); goto _3;}
        if(isLetter(ch) || isDigit(ch))
        {
          lc = lexError; lv[++i_lv]='\0'; er(3); return;
        }
        lc = lexInt; lv[++i_lv] ='\0'; 
        unum = octToInt(lv + 2);
        return;
    _4:
        if(isDigit(ch)) {lv[++i_lv]=ch; nxch(); goto _4;}
        if(isLetter(ch)) {lc = lexError; lv[++i_lv]='\0'; er(3); return;}
        lc = lexInt; lv[++i_lv] ='\0'; 
        unum = decToInt(lv + 3);
        return;
    _5:
        if(isHex(ch)) 
        {
          lv[++i_lv]=ch; nxch(); goto _5;
        }
        if(isLetter(ch)) {lc = lexError; lv[++i_lv]='\0'; er(3); return;}
        lc = lexInt; lv[++i_lv]='\0'; 
        unum = hexToInt(lv + 3);
        return;
}

//------------------------------------------------------------------------------

// Действительное, начинающееся с точки
static void IsFloatNumber(void) {
  lv[++i_lv]=ch; nxch();
  //_0:
    if(isDigit(ch)) {lv[++i_lv]=ch; nxch(); goto _1;}
    lc = lexError; lv[++i_lv]='\0'; er(5); return;
  _1:
    if(isDigit(ch)) {lv[++i_lv]=ch; nxch(); goto _1;}
    if(ch == 'E' || ch == 'e') {IsExp(); return;}
    if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return;}
    lc=lexFloat; lv[++i_lv]='\0'; 
    dnum = atof(lv);
    return; 
}

//------------------------------------------------------------------------------

// Строка символов
static void IsString(void)
{
_1:
  if(ch == '\"') { // пустиая строка или апостроф в строке
    nxch(); goto _2;
  }
  if(ch == EOF) { // конец файла, а строка не закончилась
    lc = lexError; lv[++i_lv]='\0'; er(6); return; 
  } 
  lv[++i_lv]=ch; 
  nxch();
  goto _1; // любой другой символ внутри строки
_2:
  if(ch == '\"') { // дубль для апострофа
    lv[++i_lv]=ch; // фиксируем его в строке как один     
    nxch(); 
    goto _1;
  }
  lc = lexString; lv[++i_lv] = '\0'; return; // законченная строка
}

//------------------------------------------------------------------------------
// Функция, формирующая следующую лексему
// Вызывается синтаксическим анализатором
//------------------------------------------------------------------------------

void nxl(void) {
  do {
    i_lv = -1;
    lv[0] = '\0';
    if(ch == EOF) {lc = lexEof;}
    else if(isSkip(ch)) {nxch(); lc = lexSkip;}
    else if(isLetter(ch) || ch == '_'){
      lv[++i_lv]=ch; nxch(); IsContinueIdOrKw();
    }
    else if(isDigit(ch)) {IsNumber();}
    else if(isIgnore(ch)) {nxch(); lc = lexIgnore;}
    else if(ch == '/') {nxch(); IsSlashComment();}
    else if(ch == '\"') {nxch(); IsString();}
    else if(ch == ';') {nxch(); lc = lexSemicolon;}
    else if(ch == ',') {nxch(); lc = lexComma;}
    else if(ch == ':') {
      nxch(); 
      if(ch == '=') {nxch(); lc = lexAssign;}
      else lc = lexColon;
    }
    else if(ch == '(') {nxch(); lc = lexLftRndBr;}
    else if(ch == ')') {nxch(); lc = lexRghRndBr;}
    else if(ch == '[') {nxch(); lc = lexLftSqBr;}
    else if(ch == ']') {nxch(); lc = lexRghSqBr;}
    else if(ch == '*') {nxch(); lc = lexStar;}
    else if(ch == '%') {nxch(); lc = lexPercent;}
    else if(ch == '+') {nxch(); lc = lexPlus;}
    else if(ch == '-') {
      nxch(); 
      if(ch == '>') {nxch(); lc = lexArrow;}  
      else lc = lexMinus;
    }
    else if(ch == '=') {nxch(); lc = lexEQ;}
    else if(ch == '!') {
      nxch(); 
      if(ch == '=') {nxch(); lc = lexNE;}
      else {lc = lexError; er(1);}
    }
    else if(ch == '>') {
      nxch(); 
      if(ch == '=') {nxch(); lc = lexGE;}
      lc = lexGT;
    }
    else if(ch == '<') {
      nxch();
      if(ch == '=') {nxch(); lc = lexLE;}
      lc = lexLT;
    }
    else if(ch == '{') {nxch(); IsPrefNumber();}
    else if(ch == '.') {lv[++i_lv]=ch; nxch(); IsFloatNumber();}
    else {lc = lexError; er(0); nxch();}
  } while (lc == lexComment || lc == lexSkip || lc == lexIgnore);
}

//------------------------------------------------------------------------------
// Функция, инициирующая работу со сканером
// Вызывается перед запуском синтаксического анализатора
//------------------------------------------------------------------------------

bool scan_init(char *file_name) {
  if((infil = fopen(file_name, "r"))==0) {
    printf("\nInput file is absent!\n"); return false;
  }

  nxch();
  nxl();

  return true;
}

//------------------------------------------------------------------------------
// Функция, завершающая работу сканера
//------------------------------------------------------------------------------

void scan_destroy() {
  fclose(infil);
}
