//------------------------------------------------------------------------------
// saner.cpp - модуль лексического анализатора
//------------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "scanner_data.h"
#include "scanner_local.h"

//------------------------------------------------------------------------------

void er(int i); // Импорт интерфейса функции вывода ошибок

//------------------------------------------------------------------------------
// Функции транслитератора, используемые для определения класса лексем
//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу букв
bool inline isLetter(int ch) {
  if((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу двоичных цифр
bool inline isBin(int ch) {
  if((ch == '0' || ch == '1'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу восьмиричных цифр
bool inline isOctal(int ch) {
  if((ch >= '0' && ch <= '7'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу десятичных цифр
bool inline isDigit(int ch) {
  if((ch >= '0' && ch <= '9'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу шестнадцатиричных цифр
bool inline isHex(int ch) {
  if((ch >= '0' && ch <= '9') ||
     (ch >= 'A' && ch <= 'F') || 
     (ch >= 'a' && ch <= 'f'))
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу пробелов
bool inline isSkip(int ch) {
  if(ch == ' ' || ch == '\t' || ch == '\n' || ch == '\f')
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Определяет принадлежность символа к классу игнорируемых
bool inline isIgnore(int ch) {
  if(ch >0 && ch < ' ' && ch != '\t' && ch != '\n' && ch != '\f')
    return true;
  else
    return false;
}

//------------------------------------------------------------------------------

// Читает следующий символ из входного потока
static void nxch(void)
{
    if((ch = getc(infil)) == '\n') {++line; column = 0;}
    else ++column;
    ++poz; // Переход к следующей позиции в файле
}

//------------------------------------------------------------------------------

// откат назад при неудачной попытке распознать лексему
static void unset() {
  fseek(infil, oldpoz, 0);  
  nxch();
  i_lv=-1; 
  lv[0]='\0'; 
  poz = oldpoz;
  line=oldline; 
  column=oldcolumn;
}

//------------------------------------------------------------------------------
// Максимальное значение двоичного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uBinMax = 0x3FFFFFFF; // 30 двоичных разрядов
//------------------------------------------------------------------------------

// Преобразование строки двоичных цифр в целое число
unsigned int binToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    unsigned int digit = static_cast<unsigned int>(*str);
    if(uBinMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    num <<= 1;
    num += digit - static_cast<unsigned int>('0');
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------
// Максимальное значение восьмеричного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uOctMax = 0x0FFFFFFF; // 28 двоичных разрядов
//------------------------------------------------------------------------------

// Преобразование строки восьмеричных цифр в целое число
unsigned int octToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    unsigned int digit = static_cast<unsigned int>(*str);
    if(uOctMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    num = (num << 3) + (digit - static_cast<unsigned int>('0'));
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------
// Максимальное значение двоичного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uDecMax = (0x7FFFFFFF / 10); // деление на 10 максимального целого
//------------------------------------------------------------------------------

// Преобразование строки десятичных цифр в целое число
unsigned int decToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    unsigned int digit = static_cast<unsigned int>(*str);
    if(uDecMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    num = (num * 10) + (digit - static_cast<unsigned int>('0'));
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------
// Максимальное значение восьмеричного целого числа до сдвига. 
// Используется для проверки на переполнение
unsigned int uHexMax = 0x07FFFFFF; // 28 двоичных разрядов
//------------------------------------------------------------------------------

// Преобразование строки шестнадцатиричных цифр в целое число
unsigned int hexToInt(char* str)
{
  unsigned int num = 0;
  while(*str != '\0')
  {
    if(uDecMax < num)
    { 
      // Возникнет переполнение при добавлении очередной цифры
      lc = lexError;
      er(7);
      return 0;
    }
    unsigned int digit = static_cast<unsigned int>(*str);
    num = num << 4;
    if(*str >= '0' && *str <= '9')
    {
      num += (digit - static_cast<unsigned int>('0'));
    }
    else if(*str >= 'A' && *str <= 'F')
    {
      num += (digit - static_cast<unsigned int>('A')) + 10;
    }
    else //(*str >= 'a' && *str <= 'f')
    {
      num += (digit - static_cast<unsigned int>('a')) + 10;
    }
    ++str;
  }
  return num;
}

//------------------------------------------------------------------------------

// Определяет принадлежность к ключевому слову
static void FindKw(void)
{
  for(int i = 0; i < rw_size; ++i)
    if(!strcmp(rwtab[i].wrd, lv)) {
      lc = rwtab[i].lc;
      return;
    }
  lc = lexId;
}

//------------------------------------------------------------------------------

// Распознавание идентификатора или ключевого слова
static bool IsIdOrKw()
{
    if(isLetter(ch) || ch == '_') {
      lv[++i_lv]=ch; nxch();
    }
    else 
      return false;

    while(isLetter(ch) || isDigit(ch) || ch=='_') {
        lv[++i_lv] = ch;
        nxch();
    }
    lv[++i_lv] = '\0';
    FindKw(); 
    return true;
}

//------------------------------------------------------------------------------

// Строка символов
static bool IsString(void)
{
//_0:
  if(ch == '\"') { // начало строки
    nxch(); goto _1;
  }
  else 
    return false;
_1:
  if(ch == '\"') { // пустиая строка или апостроф в строке
    nxch(); goto _2;
  }
  if(ch == EOF) { // конец файла, а строка не закончилась
    lc = lexError; 
    lv[++i_lv]='\0'; 
    er(6); 
    return true; //  чтобы не было отката
  } 
  lv[++i_lv]=ch; 
  nxch();
  goto _1; // любой другой символ внутри строки
_2:
  if(ch == '\"') { // дубль для апострофа
    lv[++i_lv]=ch; // фиксируем его в строке как один     
    nxch(); 
    goto _1;
  }
  lc = lexString; lv[++i_lv] = '\0'; return true; // законченная строка
}

//------------------------------------------------------------------------------

// Распознавание комментария
static bool IsComment()
{
    // Комментарий может оказаться длинне буфера, поэтому его текст не фиксируем
    if(ch=='/') {nxch();}
    else return false;
    if(ch != '*') {return false;}
    nxch();
  loop:
    if(ch == EOF) 
    {
      lc = lexError; 
      lv[++i_lv]='\0'; 
      er(2); 
      return true;
    } 
    while(ch != '*') 
    {
        //lv[++i_lv] = ch;
        if(ch == EOF) 
        {
          lc = lexError; 
          lv[++i_lv]='\0'; 
          er(2); return true;
        } 
        nxch();
    }
    nxch(); 
    if(ch != '/') 
    {
      //lv[++i_lv] = ch; 
      goto loop;
    }
    lv[++i_lv] = '\0';
    lc = lexComment;
    nxch();
    return true;   
}

//------------------------------------------------------------------------------

// Распознавание целого десятичного числа
static bool IsDecimal() {
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexInt; lv[++i_lv] = '\0'; 
  unum = decToInt(lv);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание целого десятичного числа с префиксом
static bool IsPrefDecimal() {
  if(ch=='{') nxch(); else return false;
  if(ch=='1'){lv[++i_lv]='1'; nxch();} else return false;
  if(ch=='0'){lv[++i_lv]='0'; nxch();} else return false;
  if(ch=='}'){lv[++i_lv]='#'; nxch();} else return false;
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexInt; lv[++i_lv] = '\0'; 
  unum = decToInt(lv + 3);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание целого двоичного числа
static bool IsBinary() {
  if(ch=='{') nxch(); else return false;
  if(ch=='2'){lv[++i_lv]='2'; nxch();} else return false;
  if(ch=='}'){lv[++i_lv]='#'; nxch();} else return false;
  if(ch=='0'||ch=='1'){lv[++i_lv]=ch; nxch();} else return false;
  while (ch=='0'||ch=='1'){
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexInt; lv[++i_lv] = '\0'; 
  unum = binToInt(lv + 2);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание целого восьмиричного числа
static bool IsOctal() {
  if(ch=='{') nxch(); else return false;
  if(ch=='8'){lv[++i_lv]='8'; nxch();} else return false;
  if(ch=='}'){lv[++i_lv]='#'; nxch();} else return false;
  if(ch>='0' && ch<='7'){lv[++i_lv]=ch; nxch();} else return false;
  while (ch>='0' && ch<='7') {
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexInt; lv[++i_lv] = '\0'; 
  unum = octToInt(lv + 2);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание целого шестнадцатиричного числа
static bool IsHex() {
  if(ch=='{') nxch(); else return false;
  if(ch=='1'){lv[++i_lv]='1'; nxch();} else return false;
  if(ch=='6'){lv[++i_lv]='6'; nxch();} else return false;
  if(ch=='}'){lv[++i_lv]='#'; nxch();} else return false;
  if(isHex(ch)) {
    lv[++i_lv]=ch; nxch();
  } 
  else return false;
  while (isHex(ch)) {
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexInt; lv[++i_lv] = '\0'; 
  unum = hexToInt(lv + 3);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание первой версии дейстительного числа,
// Состоящего из целой части и обязательного порядка
static bool IsFloat1() {
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(ch=='e'||ch=='E') {lv[++i_lv]=ch; nxch();} else return false;
  if(ch=='+'||ch=='-') {lv[++i_lv]=ch; nxch();} 
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch) || ch=='_') 
  {
    lc=lexError; lv[++i_lv]='\0'; er(3); return true;
  }
  lc = lexFloat; 
  lv[++i_lv] = '\0'; 
  dnum = atof(lv);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание второй версии дейстительного числа
static bool IsFloat2() {
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(ch=='.') {lv[++i_lv]=ch; nxch();} else return false;
  while(isDigit(ch)) {lv[++i_lv]=ch; nxch();}
  if(ch=='e'||ch=='E') {lv[++i_lv]=ch; nxch();} else return false;
  if(ch=='+'||ch=='-') {lv[++i_lv]=ch; nxch();} 
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexFloat; 
  lv[++i_lv] = '\0'; 
  dnum = atof(lv);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание третьей версии дейстительного числа
static bool IsFloat3() {
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(ch=='.') {lv[++i_lv]=ch; nxch();} else return false;
  while(isDigit(ch)) {lv[++i_lv]=ch; nxch();}
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexFloat; 
  lv[++i_lv] = '\0'; 
  dnum = atof(lv);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание четвертой версии дейстительного числа
static bool IsFloat4() {
  if(ch=='.') {lv[++i_lv]=ch; nxch();} else return false;
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(ch=='e'||ch=='E') {lv[++i_lv]=ch; nxch();} else return false;
  if(ch=='+'||ch=='-') {lv[++i_lv]=ch; nxch();} 
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexFloat; 
  lv[++i_lv] = '\0'; 
  dnum = atof(lv);
  return true; 
}

//------------------------------------------------------------------------------

// Распознавание пятой версии дейстительного числа
static bool IsFloat5() {
  if(ch=='.') {lv[++i_lv]=ch; nxch();} else return false;
  if(isDigit(ch)){lv[++i_lv]=ch; nxch();} else return false;
  while (isDigit(ch)){
    lv[++i_lv] = ch;
    nxch();
  }
  if(isLetter(ch)) {lc=lexError; lv[++i_lv]='\0'; er(3); return true;}
  lc = lexFloat; 
  lv[++i_lv] = '\0'; 
  dnum = atof(lv);
  return true; 
}

//------------------------------------------------------------------------------
// Функция, формирующая следующую лексему
// Вызывается синтаксическим анализатором
//------------------------------------------------------------------------------

void nxl(void) {
  do {
    i_lv = -1;   
    lv[0] = '\0';

    // Фиксируем начальную позицию
    oldpoz=ftell(infil)-1;
    oldline=line; oldcolumn=column;

    // Процесс пошел
    if(ch == EOF) {lc = lexEof; return;}
    // Игнорируемую лексему не возвращаем
    if(isSkip(ch)) {nxch(); lc = lexSkip; continue; /*return;*/}
    if(IsIdOrKw())       {return;} unset();
    if(IsString()) {return;} unset();
    if(IsFloat1())       {return;} unset();
    if(IsFloat2())       {return;} unset();
    if(IsFloat3())       {return;} unset();
    if(IsFloat4())       {return;} unset();
    if(IsFloat5())       {return;} unset();
    if(IsBinary())       {return;} unset();
    if(IsOctal())        {return;} unset();
    if(IsHex())          {return;} unset();
    if(IsPrefDecimal())     {return;} unset();
    if(IsDecimal())      {return;} unset();
    // Игнорируемую лексему не возвращаем
    if(IsComment())      {continue; /*return;*/} unset();
    // Игнорируемую лексему не возвращаем
    if(isIgnore(ch)) {nxch(); lc = lexIgnore; continue; /*return;*/}
    if(ch=='/') {nxch(); lc = lexSlash;return;}
    if(ch == ';') {nxch(); lc = lexSemicolon; return;}
    if(ch == ',') {nxch(); lc = lexComma; return;}
    if(ch == ':') {
      nxch(); 
      if(ch == '=') {nxch(); lc = lexAssign; return;}
    } unset();
    if(ch==':') {nxch(); lc = lexColon; return;}
    if(ch == '(') {nxch(); lc = lexLftRndBr; return;}
    if(ch == ')') {nxch(); lc = lexRghRndBr; return;}
    if(ch == '[') {nxch(); lc = lexLftSqBr; return;}
    if(ch == ']') {nxch(); lc = lexRghSqBr; return;}
    if(ch == '*') {nxch(); lc = lexStar; return;}
    if(ch == '%') {nxch(); lc = lexPercent; return;}
    if(ch == '+') {nxch(); lc = lexPlus; return;}
    if(ch == '-') {
      nxch(); 
      if(ch == '>') {nxch(); lc = lexArrow; return;}  
    } unset();
    if(ch=='-') {nxch(); lc=lexMinus; return;}
    if(ch == '=') {nxch(); lc = lexEQ; return;}
    if(ch == '!') {
      nxch(); 
      if(ch == '=') {nxch(); lc = lexNE; return;}
    } unset();
    if(ch == '>') {
      nxch(); 
      if(ch == '=') {nxch(); lc = lexGE; return;}
    } unset();
    if(ch=='>') {nxch(); lc=lexGT; return;}
    if(ch == '<') {
      nxch(); 
      if(ch == '=') {nxch(); lc = lexLE; return;}
    } unset();
    if(ch=='<') {nxch(); lc=lexLT; return;}
    lc = lexError; er(0); nxch();
  } while (lc == lexComment || lc == lexSkip || lc == lexIgnore);
}

//------------------------------------------------------------------------------
// Функция, инициирующая работу со сканером
// Вызывается перед запуском синтаксического анализатора
//------------------------------------------------------------------------------

bool scan_init(char *file_name) {
  if((infil = fopen(file_name, "r"))==0) {
    printf("\nInput file is absent!\n"); return false;
  }

  nxch();
  nxl();

  return true;
}

//------------------------------------------------------------------------------
// Функция, завершающая работу сканера
//------------------------------------------------------------------------------

void scan_destroy() {
  fclose(infil);
}
